/* PrimeQ_gen.h -- Geoffrey Reynolds, August 2009.

   This is an adaptation of Mark Rodenkirch's PPC64 mulmod code using
   only plain integer arithmetic available in C99.
*/

// Modified by Jaroslaw Wroblewski, September 3, 2009

/* Unsigned 64x64->128 bit multiply. ABH:ABL <-- A*B
 */
#define MUL128(_ABL,_ABH,_A,_B) {             \
  uint64_t _x0, _x1, _x2, _x3, _a, _b;        \
  uint32_t _al, _ah, _bl, _bh;                \
  _a = (_A); _al = _a; _ah = _a >> 32;        \
  _b = (_B); _bl = _b; _bh = _b >> 32;        \
  _x0 = (uint64_t)_al * _bl;                  \
  _x1 = (uint64_t)_al * _bh;                  \
  _x2 = (uint64_t)_ah * _bl;                  \
  _x3 = (uint64_t)_ah * _bh;                  \
  _x1 = _x1 + (_x0 >> 32) + _x2;              \
  _x3 = _x3 + ((uint64_t)(_x1 < _x2) << 32);  \
  (_ABH) = _x3 + (_x1 >> 32);                 \
  (_ABL) = (_x1 << 32) | (_x0 & 0xFFFFFFFF);  \
}

/* a*b mod p
   Assumes a,b < p < 2^63
*/
static uint64_t
 mulmod(uint64_t a, uint64_t b, uint64_t p, uint64_t magic, uint64_t shift)
{
  uint64_t ab0, ab1, mab00, mab01, mab10, mab11;
  uint64_t r, s0, s1, t;

  MUL128(ab0,ab1,a,b);
  MUL128(mab00,mab01,magic,ab0);
  MUL128(mab10,mab11,magic,ab1);

  s0 = mab01 + mab10;
  s1 = mab11 + (s0 < mab01);

  t = (s0 >> shift) | (s1 << (64-shift));

  r = ab0 - t*p;

  if ((int64_t)r < 0)
    r += p;

  return r;
}

static void getMagic(uint64_t d, uint64_t *pMagic, uint64_t *pShift)
{
   uint64_t two63 = 0x80000000;

   uint64_t t = two63;
   uint64_t p = 31;            /* Init p. */
   uint64_t q2 = two63/d;        /* Init q2 = 2**p/|d|. */
   uint64_t r2 = two63- q2*d;      /* Init r2 = rem(2**p, |d|). */
   uint64_t mag;

   do {
      p = p + 1;

      q2 = 2*q2;                  /* Update q2 = 2**p/|d|. */
      r2 = 2*r2;                  /* Update r2 = rem(2**p, |d|. */
      if (r2 >= d) {          /* (Must be an unsigned */
         q2 = q2 + 1;        /* comparison here). */
         r2 = r2 - d;
      }
   } while (p<=122);
   
// instead of 122 take 62+number_of_significant_bits_of_d
   
   mag = q2 + 1;

   *pShift = p - 64;             /* shift amount to return. */

   *pMagic = mag;
}


/* Returns 0 only if N is composite.
   Otherwise N is a strong probable prime to base a.
 */
static int strong_prp(int64_t a, int64_t N)
{
  uint64_t magic, shift;
  int64_t r;
  uint64_t d;
  unsigned int s, t;

  /* If N is prime and N = d*2^t+1, where d is odd, then either
     1.  a^d = 1 (mod N), or
     2.  a^(d*2^s) = -1 (mod N) for some s in 0 <= s < t */

  getMagic(N,&magic,&shift);

//#ifdef __GNUC__
//  t = __builtin_ctzll(N-1);
//  d = (uint64_t)N >> t;
//#else
  for (d = (uint64_t)N >> 1, t = 1; !(d & 1); d >>= 1, t++)
    ;
//#endif

  /* r <-- a^d mod N, assuming d odd */
  for (r = a, d >>= 1; d > 0; d >>= 1)
  {
    a = mulmod(a,a,N,magic,shift);
    if (d & 1)
      r = mulmod(r,a,N,magic,shift);
  }

  /* Clause 1. and s = 0 case for clause 2. */
  if (r == 1 || r == N-1)
    return 1;

  /* 0 < s < t cases for clause 2. */
  for (s = 1; s < t; s++)
    if ((r = mulmod(r,r,N,magic,shift)) == N-1)
      return 1;

  return 0;
}


/* Returns 0 only if N is composite.
   Otherwise N is a strong probable prime to base 2.
   For AP26, N can be assumed to have no prime divisors <= 541?
 */
int PrimeQ(int64_t N)
{
  return strong_prp(2,N);
}
